/* search.c is part of VIP program
 * Written by J. Salvador Arias, 2010
 *      CONICET, INSUE, Instituto Miguel Lillo,
 *      Miguel Lillo 205, S.M. de Tucuman (4000)
 *      Tucuman, Argentina
 *      e-mail: jsalarias@csnat.unt.edu.ar
 *
 * VIP is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation and
 * Creative Commons.
 *
 * This file and the VIP program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * To view a copy of this license, visit
 * http://creativecommons.org/licenses/GPL/2.0/
 * or send a letter to
 * Creative Commons, 171 Second Street, Suite 300,
 * San Francisco, California, 94105, USA.
 */

#include "biogeo.h"
#include "bitfield.h"
#include "grid.h"
#include "tessellation.h"

#define JUST_COUNT_COST     1
#define FULL_OPTIMIZATION   2
#define SEARCH_OPTIMIZATION 3

/* Code specificly used for searches and optimizations */

static SREAL OptimizeNode (GNode* node, SFLAG mode);
static void OptimizeAll (GPtrArray* recArray, SFLAG optimize);

static void CopyRec (RECONS* dest, RECONS* src);
static GPtrArray* CreateRecArrayCopy (GPtrArray* src);
static void CopyRecArray (GPtrArray* dest, GPtrArray* src);

static SFLAG CompareRec (GPtrArray* probe, GList* recList);
static SFLAG CheckReconstruction (PHYLOTREE* tree, GPtrArray* recArray, SREAL* globalBest, SCOUNTER* stored, SCOUNTER* bestHits);

static SREAL IncrementalDownPass (GNode* startNode);

static GNode* FindOverlap (GNode* node);
static SFLAG QuickOverlap (GNode* node);
static SFLAG CarefulOverlap (GNode* node);

static void MeasureOverlap (RECONS* actRec);

static void NodeRemoval (PHYLOTREE* actTree, GPtrArray* sectorNodeArray);
static void BlindAnnealing (PHYLOTREE* actTree, GPtrArray* recArray);
static void GibbsAnnealing (PHYLOTREE* actTree, GPtrArray* recArray, GPtrArray* sectorArray, INTARRAY* prob);

static void UpdateRecCounts (GPtrArray* recArray);

static BITFIELD* tmpBitField = NULL;
static BITFIELD* firstLocalBF = NULL;
static BITFIELD* secondLocalBF = NULL;

static GPtrArray* tmpNodeArray = NULL;

void SetRec (GPtrArray* recArray) {
 SCOUNTER i;
 RECONS* actRec;

 for (i = recArray -> len - 1; i >= 0; -- i) {
   actRec = (RECONS*) g_ptr_array_index (recArray, i);
   actRec -> node -> rec = actRec;
 }
}

void MakeOptimization (SFLAG justCount) {
 GList* navTree;
 GList* navRec;
 PHYLOTREE* tree;
 GNode* navNode;
 PHYLONODE* actNode;
 RECONS* actRec;
 GPtrArray* rec;

 firstLocalBF = CreateBitField ();
 for (navTree = trees; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   if (tree -> recOR == NULL) {
     /* There are no reconstruction, first creates the array */
     tree -> recOR = g_ptr_array_new ();
     for (navNode = tree -> root; navNode != NULL; ) {
       actNode = (PHYLONODE*) navNode -> data;
       if (actNode -> rec == NULL) {
         actRec = CreateRecons (actNode);
         g_ptr_array_add (tree -> recOR, actRec);
         actNode -> rec = actRec;
         if (g_node_first_child (navNode) != NULL) {
           navNode = g_node_first_child (navNode);
           continue;
         }
       }
       if (g_node_next_sibling (navNode) == NULL)
         navNode = navNode -> parent;
       else
         navNode = g_node_next_sibling (navNode);
     }
     OptimizeAll (tree -> recOR, TRUE);
   }
   else {
     /* If there are a reconstruction, updates the costs of the reconstructions */
     if (tree -> recs != NULL) {
       for (navRec = tree -> recs; navRec != NULL; navRec = g_list_next (navRec)) {
         rec = (GPtrArray*) navRec -> data;
         SetRec (rec);
         if (justCount == TRUE)
           OptimizeAll (rec, FALSE);
         else
           OptimizeAll (rec, TRUE);
       }
     }
     SetRec (tree -> recOR);
     if (justCount == TRUE)
       OptimizeAll (tree -> recOR, FALSE);
     else
       OptimizeAll (tree -> recOR, TRUE);
   }
 }
 g_slice_free1 (GETBITFIELDSIZE, firstLocalBF);
}

SREAL OptimizeNode (GNode* node, SFLAG mode) {
 RECONS* actRec;
 PHYLONODE* actNode;
 PHYLONODE* descNode;
 PHYLONODE* left;
 PHYLONODE* right;
 GNode* navDesc;
 SCOUNTER ageDif;
 SREAL overlap, divLeft, divRight, timeFraq;
 SFLAG fullRemov;

 actNode = (PHYLONODE*) node -> data;
 actRec = actNode -> rec;
 actRec -> cost = 0;
 actRec -> numVics = actRec -> numNoVics = actRec -> numJnodes = 0;
 actRec -> nodeFlag &= J_NODE;        /* If the node is a J-node, then nodeFlag is equal to J_NODE, otherwise nodeFlag is 0 */
 if (actRec -> triangles != NULL)
   DestroyTriangleList (actRec -> triangles);
 actRec -> triangles = NULL;
 if (g_node_first_child (actNode -> node) == NULL) {
   if (mode == FULL_OPTIMIZATION) {
     CopyBitField (actRec -> obs, actNode -> term -> obs);
     if (CountBits (actRec -> obs) > 0) {
       DoFill (actRec -> cells, actRec -> obs, actRec -> node -> fill);
       actRec -> numCells = CountBits (actRec -> cells);
       FillWithReference (actRec -> inher, actRec -> obs, actRec -> remov, actRec -> node -> fill, FALSE, TRUE);
       actRec -> numInher = CountBits (actRec -> inher);
     }
     else {         /* An empty set "optimization" */
       ResetBitField (actRec -> cells);
       ResetBitField (actRec -> inher);
       ResetBitField (actRec -> remov);
       actRec -> numCells = actRec -> numInher = actRec -> numRemov = 0;
     }
   }
   else if (mode == SEARCH_OPTIMIZATION) {
     ResetBitField (actRec -> inher);
     BitFieldAddNotRemoved (actRec -> inher, actRec -> cells, actRec -> remov);
     actRec -> numInher = CountBits (actRec -> inher);
   }
   if (actRec -> numCells == 0) return 0;   /* Empty set exit */
   if (actRec -> nodeFlag & J_NODE) {
     actRec -> numJnodes = 1;
     if (actRec -> numInher > 0) {
       if (params . mode & OVER_REAL) {
         overlap = actRec -> numInher / actRec -> numCells;
         actRec -> cost += (overlap * params . costPartRemov);
       }
       else
         actRec -> cost += params . costPartRemov;
     }
     else
       actRec -> cost = params . costRemov;
   }
   return actRec -> cost;
 }
 if ((actRec -> numInher == 0) && (actRec -> numRemov > 0))
   fullRemov = TRUE;
 else
   fullRemov = FALSE;
 if (mode & (FULL_OPTIMIZATION | SEARCH_OPTIMIZATION)) {
   ResetBitField (actRec -> obs);
   ResetBitField (actRec -> cells);
   actRec -> numInher = 0;
   actRec -> numRemov = 0;
   actRec -> active = 0;
   actRec -> setA = NULL;
   actRec -> setB = NULL;
 }
 left = right = NULL;
 ageDif = 0;
 for (navDesc = g_node_first_child (actNode -> node); navDesc != NULL; navDesc = g_node_next_sibling (navDesc)) {
   descNode = (PHYLONODE*) navDesc -> data;
   actRec -> numVics += descNode -> rec -> numVics;
   actRec -> numNoVics += descNode -> rec -> numNoVics;
   actRec -> numJnodes += descNode -> rec -> numJnodes;
   actRec -> cost += descNode -> rec -> cost;
   if (mode & (FULL_OPTIMIZATION | SEARCH_OPTIMIZATION)) {
     if (descNode -> rec -> numInher > 0) {
       BitFieldAddNotRemoved (actRec -> obs, descNode -> rec -> obs, descNode -> rec -> remov);
       BitFieldORsum (actRec -> cells, descNode -> rec -> inher);
       ++ actRec -> active;
       if (left == NULL) {
         left = descNode;
         ageDif = actNode -> minAge - descNode -> minAge;
       }
       else if (right == NULL) {
         right = descNode;
         if ((actNode -> minAge - descNode -> minAge) > ageDif)
           ageDif = actNode -> minAge - descNode -> minAge;
       }
     }
   }
 }
 if (mode & (FULL_OPTIMIZATION | SEARCH_OPTIMIZATION)) {
   actRec -> numCells = CountBits (actRec -> cells);
   actRec -> numRemov = CountBits (actRec -> remov);
   if (actRec -> numCells == 0) {
     /* This is an empty set */
     ResetBitField (actRec -> remov);
     actRec -> numInher = actRec -> numRemov = 0;
     actRec -> cost = 0;
     return actRec -> cost;
   }
   if (fullRemov) {
     ResetBitField (actRec -> inher);
     CopyBitField (actRec -> remov, actRec -> cells);
     actRec -> numInher = 0;
   }
   else {
     CopyBitField (actRec -> inher, actRec -> cells);
     BitFieldRemove (actRec -> inher, actRec -> remov);
     actRec -> numInher = CountBits (actRec -> inher);
   }
 }
 if (actRec -> nodeFlag & J_NODE) {
   ++ actRec -> numJnodes;
   if (actRec -> numInher > 0) {
     if (params . mode & OVER_REAL) {
       overlap = actRec -> numInher / actRec -> numCells;
       actRec -> cost += (overlap * params . costPartRemov);
     }
     else
       actRec -> cost += params . costPartRemov;
   }
   else
     actRec -> cost += params . costRemov;
 }
 if (actRec -> active == 1) {
   actRec -> nodeFlag &= ~ VIC_NODE;
   if (params . costRemov == 0) {
     ++ actRec -> numNoVics;
     actRec -> cost += params . costNoVic;
   }
   return actRec -> cost;
 }
 else if (actRec -> active != 2) {
   actRec -> nodeFlag &= ~ VIC_NODE;
   ++ actRec -> numNoVics;
   actRec -> cost += params . costNoVic;
   return actRec -> cost;
 }
 if (mode & (FULL_OPTIMIZATION | SEARCH_OPTIMIZATION)) {
   actRec -> timeLag = ageDif;
   /* Preserves the order, indep of tree sorting */
   if (left -> id < right -> id) {
     actRec -> setA = left;
     actRec -> setB = right;
   }
   else {
     actRec -> setB = left;
     actRec -> setA = right;
   }
   MeasureOverlap (actRec);
 }
 divLeft = (SREAL) actRec -> overlap / actRec -> setA -> rec -> numInher;
 divRight = (SREAL) actRec -> overlap / actRec -> setB -> rec -> numInher;
 overlap = (divLeft > divRight) ? divLeft : divRight;
 if ((overlap <= params . overlap) && (overlap >= 0)) {
   if (params . mode & USE_TIME) {
     if (actRec -> timeLag > params . maxTime) {
       actRec -> nodeFlag &= ~ VIC_NODE;
       ++ actRec -> numNoVics;
       actRec -> cost += params . costNoVic;
       return actRec -> cost;
     }
     if (actRec -> timeLag > params . timeLag) {
       if (params . mode & TIME_REAL) {
         timeFraq = (SREAL) (ageDif - params . timeLag) / params . timeDiff;
         actRec -> cost += timeFraq * params . costTimeLag;
       }
       else
         actRec -> cost += params . costTimeLag;
       actRec -> nodeFlag |= TIME_LAG;
     }
   }
   actRec -> nodeFlag |= VIC_NODE;
   ++ actRec -> numVics;
   if (params . mode & OVER_REAL)
     actRec -> cost += (overlap * params . costNoVic);
 }
 else {
   actRec -> nodeFlag &= ~ VIC_NODE;
   ++ actRec -> numNoVics;
   actRec -> cost += params . costNoVic;
 }
 return actRec -> cost;
}

void OptimizeAll (GPtrArray* recArray, SFLAG optimize) {
 SINTEGER i;
 RECONS* actRec;

 for (i = recArray -> len - 1; i >= 0; -- i) {
   actRec = (RECONS*) g_ptr_array_index (recArray, i);
   OptimizeNode (actRec -> node -> node, FULL_OPTIMIZATION);
 }
}

SREAL IncrementalDownPass (GNode* startNode) {
 GNode* navNode;
 SREAL toRetCost;

 toRetCost = G_MAXDOUBLE;
 for (navNode = startNode; navNode != NULL; navNode = navNode -> parent)
   toRetCost = OptimizeNode (navNode, SEARCH_OPTIMIZATION);
 return toRetCost;
}

GNode* FindOverlap (GNode* node) {
 GNode* prevNode;
 GNode* navNode;
 PHYLONODE* actNode;
 PHYLONODE* ancNode;
 PHYLONODE* sisNode;
 PHYLONODE* sisOfAct;
 RECONS* nodeRec;
 RECONS* ancRec;
 SCOUNTER bitCount;
 SREAL divSis, divRec, overlap;

 actNode = (PHYLONODE*) node -> data;
 nodeRec = actNode -> rec;
 if (nodeRec -> numInher == 0) return NULL;
 prevNode = node -> parent;
 if (prevNode == NULL) return NULL;
 ResetBitField (firstLocalBF);         /* track valid cells */
 CopyBitField (firstLocalBF, nodeRec -> inher);
 ancNode = (PHYLONODE*) prevNode -> data;
 if (ancNode -> rec -> setA != NULL) {
 /* This guarantees than when overlapping with the sister node, the removal "skip out" the allowed overlapp
    (that is, accept overlapps greater then params -> overlap for vicariant nodes!) */
 /* Example ([ab] ([c] ([b] [bd]))), the removal of b in [bd] must be invalid */
   sisOfAct = (ancNode -> rec -> setA != actNode) ? ancNode -> rec -> setA : ancNode -> rec -> setB;
   BitFieldRemove (firstLocalBF, sisOfAct -> rec -> inher);
 }
 BitFieldRemove (firstLocalBF, ancNode -> rec -> remov);
 for (navNode = prevNode -> parent; navNode != NULL; navNode = navNode -> parent) {
   ancNode = (PHYLONODE*) navNode -> data;
   ancRec = ancNode -> rec;
   if (ancRec -> numInher == 0) return NULL;
   if ((ancRec -> nodeFlag & VIC_NODE) || (ancRec -> active != 2)) {
     BitFieldRemove (firstLocalBF, ancRec -> remov);
     prevNode = navNode;
     continue;
   }
   sisNode = (ancNode -> rec -> setA -> node != prevNode) ? ancNode -> rec -> setA : ancNode -> rec -> setB;
   bitCount = BitFieldCommon (firstLocalBF, sisNode -> rec -> inher);
   if (bitCount > 0) {
     divSis = (SREAL) bitCount / sisNode -> rec -> numCells;
     divRec = (SREAL) (bitCount + nodeRec -> numRemov) / nodeRec -> numCells;
     overlap = (divSis > divRec) ? divSis : divRec;
     if (overlap <= params . lateOverlap) return sisNode -> node;
   }
   BitFieldRemove (firstLocalBF, ancRec -> remov);
   prevNode = navNode;
 }
 return NULL;
}

/* Just removes the cells in the overlap, successptible to "orphan" cells */
SFLAG QuickOverlap (GNode* node) {
 GNode* found;
 PHYLONODE* sisNode;
 PHYLONODE* actNode;

 found = FindOverlap (node);
 if (found == NULL) return FALSE;
 actNode = (PHYLONODE*) node -> data;
 sisNode = (PHYLONODE*) found -> data;
 /* Updates the overlap, in a "quick" (i.e. simple) way */
 BitFieldANDsum (firstLocalBF, sisNode -> rec -> inher);
 BitFieldORsum (actNode -> rec -> remov, firstLocalBF);
 CopyBitField (actNode -> rec -> inher, actNode -> rec -> cells);
 BitFieldRemove (actNode -> rec -> inher, actNode -> rec -> remov);
 actNode -> rec -> numInher = CountBits (actNode -> rec -> inher);
 actNode -> rec -> numRemov = CountBits (actNode -> rec -> remov);
 actNode -> rec -> nodeFlag |= J_NODE;
 return TRUE;
}

/* Checks the original records in the overlap zone, if a point falls in the overlap
   then its cell and their filling cells is not included in the inher cells */
SFLAG CarefulOverlap (GNode* node) {
 GNode* found;
 GNode* navNode;
 GNode* ancNode;
 PHYLONODE* sisNode;
 PHYLONODE* actNode;
 PHYLONODE* childNode;
 PHYLONODE* anc;
 TERMINAL* term;

 found = FindOverlap (node);
 if (found == NULL) return FALSE;
 actNode = (PHYLONODE*) node -> data;
 sisNode = (PHYLONODE*) found -> data;
 /* Updates the overlap cells */
 BitFieldANDsum (firstLocalBF, sisNode -> rec -> inher);
 BitFieldORsum (actNode -> rec -> remov, firstLocalBF);
 ResetBitField (actNode -> rec -> inher);
 for (navNode = node; navNode != NULL;) {
   if (IsDesc (navNode, node) != TRUE) break;
   childNode = (PHYLONODE*) navNode -> data;
   if (childNode -> rec -> numInher == 0) {
   /* A full j-node, ignores all of its descendants */
     while (navNode != NULL) {
       if (g_node_next_sibling (navNode) != NULL) {
         navNode = g_node_next_sibling (navNode);
         break;
       }
       navNode = navNode -> parent;
     }
     continue;
   }
   if (g_node_first_child (navNode) != NULL) {
   /* Not a term, goes to the first descendant */
     navNode = g_node_first_child (navNode);
     continue;
   }
   /* Updates the removed cells */
   ResetBitField (secondLocalBF);
   for (ancNode = navNode; ; ancNode = ancNode -> parent) {
     anc = (PHYLONODE*) ancNode -> data;
     BitFieldORsum (secondLocalBF, anc -> rec -> remov);
     if (ancNode == node) break;
   }
   term = childNode -> term;
   FillWithReference (actNode -> rec -> inher, term -> obs, secondLocalBF, childNode -> fill, FALSE, FALSE);
   while (navNode != NULL) {
     if (g_node_next_sibling (navNode) != NULL) {
       navNode = g_node_next_sibling (navNode);
       break;
     }
     navNode = navNode -> parent;
   }
 }
 /* Updates the cells of the reconstruction */
 BitFieldRemove (actNode -> rec -> inher, actNode -> rec -> remov);          /* Remove overlapped cells */
 /* Remove all non active cells */
 CopyBitField (actNode -> rec -> remov, actNode -> rec -> cells);
 BitFieldRemove (actNode -> rec -> remov, actNode -> rec -> inher);
 actNode -> rec -> numInher = CountBits (actNode -> rec -> inher);
 actNode -> rec -> numRemov = CountBits (actNode -> rec -> remov);
 actNode -> rec -> nodeFlag |= J_NODE;
 return TRUE;
}

void MeasureOverlap (RECONS* actRec) {
 SFLAG overA, overB;

 actRec -> overlap = actRec -> setA -> rec -> numInher + actRec -> setB -> rec -> numInher - actRec -> numCells;
 if (actRec -> overlap == 0) return;
 overA = BitFieldAnd (actRec -> setB -> rec -> inher, actRec -> setA -> rec -> obs);
 if (overA == TRUE) return;
 overB = BitFieldAnd (actRec -> setA -> rec -> inher, actRec -> setB -> rec -> obs);
 if (overB != TRUE) {
   actRec -> overlap = 0;
   return;
 }
}

void NodeRemoval (PHYLOTREE* actTree, GPtrArray* sectorNodeArray) {
 SREAL bestScore, iniScore, actScore;
 SFLAG foundBetter, foundEqual, foundOverlap;
 SFLAG removNode;
 SFLAG tmpNodeFlag;
 SFLAG update;
 SFLAG doBreak;
 GNode* navNode;
 PHYLONODE* actNode;
 PHYLONODE* tmpNode;
 RECONS* actRec;
 SCOUNTER k, bestPos;
 SCOUNTER tmpInher, tmpRemov;
 SCOUNTER randNum;

 foundBetter = TRUE;
 tmpInher = 0;
 while (foundBetter == TRUE) {
   foundBetter = FALSE;
   foundOverlap = FALSE;
   foundEqual = FALSE;
   ShufflePtrArray (sectorNodeArray);
   actNode = (PHYLONODE*) actTree -> root -> data;
   iniScore = bestScore = actNode -> rec -> cost;
   k = bestPos = 0;
   update = TRUE;
   doBreak = FALSE;
   for (k = 0; k < sectorNodeArray -> len; ++ k) {
     navNode = (GNode*) g_ptr_array_index (sectorNodeArray, k);
     if (navNode -> parent == NULL) continue;
     actNode = (PHYLONODE*) navNode -> data;
     actRec = actNode -> rec;
     if (actRec -> numCells == 0) continue;
     if ((actRec -> nodeFlag & J_NODE) && (params . algorithm & FLIP_NODES)) {
       /* Make the backup of the node */
       CopyBitField (tmpBitField, actRec -> remov);
       if (g_node_first_child (navNode) == NULL)
         tmpInher = actRec -> numInher;
       tmpRemov = actRec -> numRemov;
       tmpNodeFlag = actRec -> nodeFlag;
       /* Make the change */

       actRec -> nodeFlag &= ~ J_NODE;
       ResetBitField (actRec -> remov);
       CopyBitField (actRec -> inher, actRec -> cells);
       actRec -> numInher = actRec -> numCells;
       actRec -> numRemov = 0;
       actScore = IncrementalDownPass (navNode);
       if (actScore < bestScore) {
         foundBetter = TRUE;
         bestScore = actScore;
         bestPos = k;
         if (params . algorithm & FAST_SEARCH) {
           update = FALSE;
           break;
         }
       }
       else if ((actScore == bestScore) && (foundEqual == FALSE)) {
         foundEqual = TRUE;
         bestPos = k;
         if (params . algorithm & FAST_SEARCH) {
           randNum = g_random_int_range (0, 100);
           if (randNum < params . acceptEqual) {
             foundBetter = TRUE;
             update = FALSE;
             break;
           }
         }
       }
     }
     else if (actRec -> numInher != 0) {
       tmpNode = (PHYLONODE*) navNode -> parent -> data;
       if ((tmpNode -> rec -> numInher == 0) || (tmpNode -> rec -> active < 2)) continue;
       if (g_node_first_child (navNode) != NULL) {
         if (actRec -> active < 2) continue;
       }
       /* Make the backup of the node */
       CopyBitField (tmpBitField, actRec -> remov);
       if (g_node_first_child (navNode) == NULL)
         tmpInher = actRec -> numInher;
       tmpRemov = actRec -> numRemov;
       tmpNodeFlag = actRec -> nodeFlag;

       /* Check the node */
       randNum = 0;
       if (params . search & USE_LATEOVER)
         randNum = g_random_int_range (0, 100);
       removNode = FALSE;
       if (randNum >= 50) {
         if (QuickOverlap (navNode) == TRUE) {
           actScore = IncrementalDownPass (navNode);
           if (actScore < bestScore) {
             foundBetter = TRUE;
             foundOverlap = TRUE;
             bestScore = actScore;
             bestPos = k;
             removNode = TRUE;
             if (params . algorithm & FAST_SEARCH)
               doBreak = TRUE;
           }
           else if ((actScore == bestScore) && (foundEqual == FALSE)) {
             foundEqual = TRUE;
             foundOverlap = TRUE;
             bestPos = k;
             removNode = TRUE;
             if (params . algorithm & FAST_SEARCH) {
               randNum = g_random_int_range (0, 100);
               if (randNum < params . acceptEqual) {
                 foundBetter = TRUE;
                 doBreak = TRUE;
               }
             }
           }
         }
       }
       if ((!(actRec -> nodeFlag & J_NODE)) && (removNode != TRUE)) {
         actRec -> nodeFlag |= J_NODE;
         actRec -> numInher = 0;
         CopyBitField (actRec -> remov, actRec -> cells);
         actRec -> numRemov = CountBits (actRec -> remov);
         actScore = IncrementalDownPass (navNode);
         if (actScore < bestScore) {
           foundBetter = TRUE;
           bestScore = actScore;
           bestPos = k;
           removNode = TRUE;
           if (params . algorithm & FAST_SEARCH) {
             update = FALSE;
             break;
           }
         }
         else if ((actScore == bestScore) && (foundEqual == FALSE)) {
           foundEqual = TRUE;
           bestPos = k;
           removNode = TRUE;
           if (params . algorithm & FAST_SEARCH) {
             randNum = g_random_int_range (0, 100);
             if (randNum < params . acceptEqual) {
               foundBetter = TRUE;
               update = FALSE;
               break;
             }
           }
         }
       }
       /* Search for overlap */
       if ((params . search & USE_LATEOVER) && (removNode != TRUE) && (randNum < 50)) {
         if (QuickOverlap (navNode) == TRUE) {
           actScore = IncrementalDownPass (navNode);
           if (actScore < bestScore) {
             foundBetter = TRUE;
             foundOverlap = TRUE;
             bestScore = actScore;
             bestPos = k;
             if (params . algorithm & FAST_SEARCH)
               doBreak = TRUE;
           }
           else if ((actScore == bestScore) && (foundEqual == FALSE)) {
             foundEqual = TRUE;
             foundOverlap = TRUE;
             bestPos = k;
             if (params . algorithm & FAST_SEARCH) {
               randNum = g_random_int_range (0, 100);
               if (randNum < params . acceptEqual) {
                 foundBetter = TRUE;
                 doBreak = TRUE;
               }
             }
           }
         }
       }
     }
     else continue;
     /* Restores the reconstruction */
     actRec -> nodeFlag = tmpNodeFlag;
     CopyBitField (actRec -> remov, tmpBitField);
     actRec -> numRemov = tmpRemov;
     if (g_node_first_child (navNode) == NULL) {
       actRec -> numInher = tmpInher;
       CopyBitField (actRec -> inher, actRec -> cells);
     }
     IncrementalDownPass (navNode);
     if (doBreak == TRUE) break;
   }
   if ((foundEqual == TRUE) && (foundBetter == FALSE)) {
     randNum = g_random_int_range (0, 100);
     if (randNum < params . acceptEqual)
       foundBetter = TRUE;
   }
   if (foundBetter == TRUE) {
     if (update == TRUE) {
       navNode = g_ptr_array_index (sectorNodeArray, bestPos);
       actNode = (PHYLONODE*) navNode -> data;
       actRec = actNode -> rec;
       if (actRec -> nodeFlag & J_NODE) {
         actRec -> nodeFlag &= ~ J_NODE;
         ResetBitField (actRec -> remov);
         CopyBitField (actRec -> inher, actRec -> cells);
         actRec -> numInher = actRec -> numCells;
         actRec -> numRemov = 0;
       }
       else {
         actRec -> nodeFlag |= J_NODE;
         if (foundOverlap == TRUE)
           CarefulOverlap (navNode);
         else {
           CopyBitField (actRec -> remov, actRec -> cells);
           actRec -> numInher = 0;
         }
       }
       bestScore = IncrementalDownPass (navNode);
     }
   }
 }
}

void BlindAnnealing (PHYLOTREE* actTree, GPtrArray* recArray) {
 SREAL actScore, prevScore, diff;
 SCOUNTER actGen, toSel, object;
 SCOUNTER tmpInher, tmpRemov;
 GNode* navNode;
 PHYLONODE* actNode;
 PHYLONODE* tmpNode;
 PHYLONODE* rootNode;
 SFLAG tmpNodeFlag;
 SFLAG restore;
 RECONS* actRec;

 rootNode = (PHYLONODE*) actTree -> root -> data;
 for (actGen = 0; actGen < params . generations; ++ actGen) {
   if (actGen % params . abandon == 0)
     NodeRemoval (actTree, tmpNodeArray);
   prevScore = rootNode -> rec -> cost;
   toSel = g_random_int_range (0, tmpNodeArray -> len);
   navNode = (GNode*) g_ptr_array_index (tmpNodeArray, toSel);
   if (navNode -> parent == NULL) continue;
   /* If the node must be eliminated, check if elimination is valid */
   actNode = (PHYLONODE*) navNode -> data;
   actRec = actNode -> rec;
   if (actRec -> numCells == 0) continue;
   if (!(actRec -> nodeFlag & J_NODE)) {
     tmpNode = (PHYLONODE*) navNode -> parent -> data;
     if ((tmpNode -> rec -> numInher == 0) || (tmpNode -> rec -> active < 2)) continue;
     if (g_node_first_child (navNode) != NULL) {
       if (actRec -> active < 2) continue;
     }
   }

   /* Make the backup of the node */
   tmpInher = 0;
   CopyBitField (tmpBitField, actRec -> remov);
   if (g_node_first_child (navNode) == NULL)
     tmpInher = actRec -> numInher;
   tmpRemov = actRec -> numRemov;
   tmpNodeFlag = actRec -> nodeFlag;

   /* Make the change */
   ChangeNode (navNode);
   actScore = rootNode -> rec -> cost;
   diff = actScore - prevScore;
   restore = FALSE;
   if (diff >= 0) {
     object = (params . acceptEqual * 1000) / (diff + 1);
     toSel = g_random_int_range (0, 1000);
     if (toSel >= object)
       restore = TRUE;
   }
   if (restore == TRUE) {
     /* Restores the reconstruction */
     actRec -> nodeFlag = tmpNodeFlag;
     CopyBitField (actRec -> remov, tmpBitField);
     actRec -> numRemov = tmpRemov;
     if (g_node_first_child (navNode) == NULL) {
       actRec -> numInher = tmpInher;
       CopyBitField (actRec -> inher, actRec -> cells);
     }
     IncrementalDownPass (navNode);
   }
 }
}

void GibbsAnnealing (PHYLOTREE* actTree, GPtrArray* recArray, GPtrArray* sectorArray, INTARRAY* prob) {
 SREAL actScore, prevScore, diff;
 SCOUNTER actGen, toSel, object, sumProb, i;
 SCOUNTER tmpInher, tmpRemov;
 GNode* navNode;
 PHYLONODE* actNode;
 PHYLONODE* tmpNode;
 PHYLONODE* rootNode;
 SFLAG tmpNodeFlag;
 RECONS* actRec;

 rootNode = (PHYLONODE*) actTree -> root -> data;
 for (actGen = 0; actGen < params . generations; ++ actGen) {
   if (actGen % params . abandon == 0)
     NodeRemoval (actTree, tmpNodeArray);
   prevScore = rootNode -> rec -> cost;
   prob -> len = 0;
   sumProb = 0;
   /* Make the node list */
   if (sectorArray -> len > 0)
     g_ptr_array_remove_range (sectorArray, 0, sectorArray -> len);
   while (sectorArray -> len < params . sectSize) {
     toSel = g_random_int_range (0, tmpNodeArray -> len);
     navNode = (GNode*) g_ptr_array_index (tmpNodeArray, toSel);
     if (navNode -> parent == NULL) continue;
     /* If the node must be eliminated, check if elimination is valid */
     actNode = (PHYLONODE*) navNode -> data;
     actRec = actNode -> rec;
     if (actRec -> numCells == 0) continue;
     if (!(actRec -> nodeFlag & J_NODE)) {
       tmpNode = (PHYLONODE*) navNode -> parent -> data;
       if ((tmpNode -> rec -> numInher == 0) || (tmpNode -> rec -> active < 2)) continue;
       if (g_node_first_child (navNode) != NULL) {
         if (actRec -> active < 2) continue;
       }
     }
     g_ptr_array_add (sectorArray, navNode);
   }
   /* Remove each element on the list */
   for (i = 0; i < params . sectSize; ++ i) {
     navNode = (GNode*) g_ptr_array_index (sectorArray, i);
     actNode = (PHYLONODE*) navNode -> data;
     actRec = actNode -> rec;
     /* Make the backup of the node */
     CopyBitField (tmpBitField, actRec -> remov);
     tmpInher = 0;
     if (g_node_first_child (navNode) == NULL)
       tmpInher = actRec -> numInher;
     tmpRemov = actRec -> numRemov;
     tmpNodeFlag = actRec -> nodeFlag;
     /* Make the change */
     ChangeNode (navNode);
     actScore = rootNode -> rec -> cost;
     diff = actScore - prevScore;
     if (diff >= 0)
       object = (params . acceptEqual * 1000) / (diff + 1);
     else {
       diff *= -1;
       object = (params . acceptEqual * 1000 * diff) / (diff + 1);
     }
     AddToArray (prob, object);
     sumProb += object;
     /* Restores the reconstruction */
     actRec -> nodeFlag = tmpNodeFlag;
     CopyBitField (actRec -> remov, tmpBitField);
     actRec -> numRemov = tmpRemov;
     if (g_node_first_child (navNode) == NULL) {
       actRec -> numInher = tmpInher;
       CopyBitField (actRec -> inher, actRec -> cells);
     }
     IncrementalDownPass (navNode);
   }
   toSel = g_random_int_range (0, sumProb);
   diff = 0;
   for (i = 0; i < prob -> len; ++ i) {
     diff += GetFromArray (prob, i);
     if (diff > toSel) break;
   }
   navNode = (GNode*) g_ptr_array_index (sectorArray, i);
   ChangeNode (navNode);
 }
}

void CopyRec (RECONS* dest, RECONS* src) {
 dest -> node = src -> node;
 CopyBitField (dest -> cells, src -> cells);
 CopyBitField (dest -> remov, src -> remov);
 CopyBitField (dest -> inher, src -> inher);
 CopyBitField (dest -> obs, src -> obs);
 dest -> setA = src -> setA;
 dest -> setB = src -> setB;
 dest -> nodeFlag = src -> nodeFlag;
 dest -> numVics = src -> numVics;
 dest -> numJnodes = src -> numJnodes;
 dest -> numNoVics = src -> numNoVics;
 dest -> cost = src -> cost;
 dest -> numCells = src -> numCells;
 dest -> numRemov = src -> numRemov;
 dest -> numInher = src -> numInher;
 dest -> active = src -> active;
 dest -> overlap = src -> overlap;
 dest -> timeLag = src -> timeLag;
 dest -> triangles = NULL;
}

GPtrArray* CreateRecArrayCopy (GPtrArray* src) {
 SCOUNTER i;
 GPtrArray* dest;
 RECONS* nuRec;

 dest = g_ptr_array_sized_new (src -> len);
 for (i = 0; i < src -> len; ++ i) {
   nuRec = (RECONS*) g_slice_alloc (sizeof (RECONS));
   nuRec -> cells = CreateBitField ();
   nuRec -> remov = CreateBitField ();
   nuRec -> inher = CreateBitField ();
   nuRec -> obs = CreateBitField ();
   nuRec -> triangles = NULL;
   CopyRec (nuRec, (RECONS*) g_ptr_array_index (src, i));
   g_ptr_array_add (dest, nuRec);
 }
 return dest;
}

void CopyRecArray (GPtrArray* dest, GPtrArray* src) {
 SCOUNTER i;

 for (i = 0; i < src -> len; ++ i)
   CopyRec ((RECONS*) g_ptr_array_index (dest, i), (RECONS*) g_ptr_array_index (src, i));
}

SFLAG CompareRec (GPtrArray* probe, GList* recList) {
 GList* navList;
 GPtrArray* recArray;
 SFLAG isEqual;
 RECONS* probeRec;
 RECONS* actRec;
 SCOUNTER i;

 for (navList = recList; navList != NULL; navList = g_list_next (navList)) {
   recArray = (GPtrArray*) navList -> data;
   isEqual = TRUE;
   for (i = 0; i < probe -> len; ++ i) {
     probeRec = (RECONS*) g_ptr_array_index (probe, i);
     actRec = (RECONS*) g_ptr_array_index (recArray, i);
     if (probeRec -> cost != actRec -> cost) {
       isEqual = FALSE;
       break;
     }
     if (probeRec -> nodeFlag != actRec -> nodeFlag) {
       isEqual = FALSE;
       break;
     }
     if (probeRec -> numVics != actRec -> numVics) {
       isEqual = FALSE;
       break;
     }
     if (probeRec -> numJnodes != actRec -> numJnodes) {
       isEqual = FALSE;
       break;
     }
   }
   if (isEqual == TRUE) return TRUE;
 }
 return FALSE;
}

SFLAG CheckReconstruction (PHYLOTREE* tree, GPtrArray* recArray, SREAL* globalBest, SCOUNTER* stored, SCOUNTER* bestHits) {
 RECONS* actRec;
 SREAL actScore;
 SFLAG addToRec;

 actRec = (RECONS*) g_ptr_array_index (recArray, 0);
 actScore = actRec -> cost;
 addToRec = params . keepAll;
 if (actScore <= *globalBest) {
   if (actScore < *globalBest) {
     *globalBest = actScore;
     addToRec = TRUE;
     *bestHits = 1;
     if ((tree -> recs != NULL) && (params . keepAll == FALSE)) {
       DestroyRecList (tree -> recs);
       tree -> recs = tree -> actRec = NULL;
       tree -> stored = *stored = 0;
     }
   }
   else {
     addToRec = TRUE;
     ++ *bestHits;
   }
 }
 /* Check is the rec to store is different from stored recs */
 if ((addToRec == TRUE) && (tree -> recs != NULL)) {
   if (CompareRec (recArray, tree -> recs) == TRUE)
     addToRec = FALSE;
 }
 return addToRec;
}

void SearchMain (PROGRESSSTRUCT* progress) {
 GList* navTree;
 GList* navRec;
 GList* lastLink;
 PHYLOTREE* tree;
 GPtrArray* sectorArray;
 GPtrArray* recArray;
 GPtrArray* tmpRec;
 RECONS* actRec;
 SREAL globalBest;
 SCOUNTER stored, bestHits;
 SCOUNTER i, j, k;
 SFLAG addToRec;
 INTARRAY* prob;

 /* Start temporal bitfields */
 tmpBitField = CreateBitField ();
 firstLocalBF = CreateBitField ();
 secondLocalBF = CreateBitField ();

 progress -> actBar1 = j = 0;
 prob = NULL;
 if (params . algorithm & USE_SECTORS)
   sectorArray = g_ptr_array_sized_new (params . sectSize);
 else
   sectorArray = NULL;
 if (params . algorithm & GIBBS_ANNEAL) {
   if (sectorArray == NULL)
     sectorArray = g_ptr_array_sized_new (params . sectSize);
   prob = CreateIntArray (params . sectSize);
 }
 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   if (progress -> isActive == FALSE) {
     g_string_append_printf (outLog,"\n\n**** Search interrupted ****");
     break;
   }
   tree = (PHYLOTREE*) navTree -> data;
   tmpNodeArray = g_ptr_array_sized_new (tree -> nodeArray -> len);
   actRec = (RECONS*) g_ptr_array_index (tree -> recOR, 0);
   globalBest = 2 * actRec -> cost;
   lastLink = NULL;
   if (tree -> recs != NULL) {
     for (navRec = tree -> recs; navRec != NULL; navRec = g_list_next (navRec)) {
       tmpRec = (GPtrArray*) navRec -> data;
       actRec = (RECONS*) g_ptr_array_index (tmpRec, 0);
       if (actRec -> cost < globalBest)
         globalBest = actRec -> cost;
       lastLink = navRec;
     }
   }
   stored = 0;
   bestHits = 0;
   recArray = NULL;
   for (i = 0; i < params . its; ++ i) {
     if (progress -> isActive == FALSE) break;
     progress -> actBar2 = i;
     UpdateOnCalculus ();
     if (tmpNodeArray -> len > 0)
       g_ptr_array_remove_range (tmpNodeArray, 0, tmpNodeArray -> len);
     CopyPtrArray (tmpNodeArray, tree -> nodeArray);
     if (recArray == NULL)
       recArray = CreateRecArrayCopy (tree -> recOR);
     else
       CopyRecArray (recArray, tree -> recOR);
     SetRec (recArray);
     if (params . algorithm & USE_SECTORS) {
       ShufflePtrArray (tmpNodeArray);
       for (k = 0; k < tmpNodeArray -> len; k += params . sectSize) {
         if (sectorArray -> len > 0)
           g_ptr_array_remove_range (sectorArray, 0, sectorArray -> len);
         CopyPtrArrayPart (sectorArray, tmpNodeArray, k, k + params . sectSize);
         NodeRemoval (tree, sectorArray);
         if (params . algorithm & FIRST_SECTOR) break;
       }
     }
     if (params . algorithm & BLIND_ANNEAL)
       BlindAnnealing (tree, recArray);
     else if (params . algorithm & GIBBS_ANNEAL)
       GibbsAnnealing (tree, recArray, sectorArray, prob);
     NodeRemoval (tree, tmpNodeArray);
     addToRec = CheckReconstruction (tree, recArray, &globalBest, &stored, &bestHits);
     if (tree -> recs == NULL)
       lastLink = NULL;
     /* Add the new reconstruction */
     if (addToRec == TRUE) {
       lastLink = InsertListElement (lastLink, recArray);
       if (tree -> recs == NULL)
         tree -> recs = lastLink;
       recArray = NULL;
       ++ stored;
     }
   }
   g_string_append_printf (outLog,"\n\nTree %ld (%s)\nBest score: %0.4f\nHits: %ld of %ld iterations\nStored reconstructions: %ld",
                                   j, tree -> name, globalBest, bestHits, i, stored);
   tree -> stored += stored;
   if (recArray != NULL)
     g_ptr_array_free (recArray, TRUE);
   SetRec (tree -> recOR);
   tree -> actRec = NULL;
   g_ptr_array_free (tmpNodeArray, TRUE);
   ++ j;
   progress -> actBar1 = j;
 }
 if (sectorArray != NULL)
   g_ptr_array_free (sectorArray, TRUE);
 if (prob != NULL)
   DestroyIntArray (prob);
 g_slice_free1 (GETBITFIELDSIZE, tmpBitField);
 g_slice_free1 (GETBITFIELDSIZE, firstLocalBF);
 g_slice_free1 (GETBITFIELDSIZE, secondLocalBF);
}

void SetNodeAsRemoved (GNode* node) {
 PHYLONODE* actNode;
 RECONS* actRec;

 actNode = (PHYLONODE*) node -> data;
 actRec = actNode -> rec;
 actRec -> nodeFlag |= J_NODE;
 ResetBitField (actRec -> inher);
 CopyBitField (actRec -> remov, actRec -> cells);
 actRec -> numInher = 0;
 actRec -> numRemov = actRec -> numCells;
 IncrementalDownPass (node);
}

void ChangeNode (GNode* node) {
 PHYLONODE* actNode;
 RECONS* actRec;

 actNode = (PHYLONODE*) node -> data;
 actRec = actNode -> rec;
 if (actRec -> nodeFlag & J_NODE) {
   actRec -> nodeFlag &= ~ J_NODE;
   ResetBitField (actRec -> remov);
   CopyBitField (actRec -> inher, actRec -> cells);
   actRec -> numInher = actRec -> numCells;
   actRec -> numRemov = 0;
   IncrementalDownPass (node);
 }
 else
   SetNodeAsRemoved (node);
}

void UpdateRecCounts (GPtrArray* recArray) {
 PHYLONODE* actNode;
 PHYLONODE* descNode;
 RECONS* actRec;
 GNode* navDesc;
 SCOUNTER i;

 for (i = recArray -> len - 1; i >= 0; -- i) {
   actRec = (RECONS*) g_ptr_array_index (recArray, i);
   actNode = actRec -> node;
   actRec -> cost = 0;
   actRec -> numVics = actRec -> numNoVics = 0;
   if (g_node_first_child (actNode -> node == NULL)) continue;
   for (navDesc = g_node_first_child (actNode -> node); navDesc != NULL; navDesc = g_node_next_sibling (navDesc)) {
     descNode = (PHYLONODE*) navDesc -> data;
     actRec -> numVics += descNode -> rec -> numVics;
     actRec -> numNoVics += descNode -> rec -> numNoVics;
   }
   if (actRec -> nodeFlag & VIC_NODE)
     ++ actRec -> numVics;
   else
     ++ actRec -> numNoVics;
 }
}

void MakeConsensus () {
 GList* navTree;
 GList* navRec;
 GPtrArray* tmpRec;
 SCOUNTER i;
 PHYLOTREE* tree;
 RECONS* actRec;
 RECONS* srcRec;

 for (navTree = GETTREES; navTree != NULL; navTree = g_list_next (navTree)) {
   tree = (PHYLOTREE*) navTree -> data;
   if (tree -> recCons != NULL) {
     SetRec (tree -> recCons);
     continue;
   }
   tmpRec = (GPtrArray*) tree -> recs -> data;
   tree -> recCons = CreateRecArrayCopy (tmpRec);
   for (navRec = g_list_next (tree -> recs); navRec != NULL; navRec = g_list_next (navRec)) {
     tmpRec = (GPtrArray*) navRec -> data;
     for (i = 0; i < tmpRec -> len; ++ i) {
       actRec = (RECONS*) g_ptr_array_index (tree -> recCons, i);
       srcRec = (RECONS*) g_ptr_array_index (tmpRec, i);
       BitFieldANDsum (actRec -> cells, srcRec -> cells);
       BitFieldANDsum (actRec -> inher, srcRec -> inher);
       BitFieldANDsum (actRec -> remov, srcRec -> remov);
       BitFieldANDsum (actRec -> obs, srcRec -> obs);
       if ((actRec -> setA != srcRec -> setA) || (actRec -> setB != srcRec -> setB)) {
         actRec -> setA = NULL;
         actRec -> setB = NULL;
         actRec -> nodeFlag &= ~ VIC_NODE;
       }
       if ((actRec -> nodeFlag & VIC_NODE) && (!(srcRec -> nodeFlag & VIC_NODE)))
         actRec -> nodeFlag &= ~ VIC_NODE;
       if ((actRec -> nodeFlag & J_NODE) && (!(srcRec -> nodeFlag & J_NODE)))
         actRec -> nodeFlag &= ~ J_NODE;
       if (actRec -> numJnodes > srcRec -> numJnodes)
         actRec -> numJnodes = srcRec -> numJnodes;
     }
   }
   tree -> suppNodes = 0;
   SetRec (tree -> recCons);
   UpdateRecCounts (tree -> recCons);
 }
}

GList* CopyActRec (PHYLOTREE* tree) {
 GPtrArray* recArray;
 GList* toRet;

 if (tree -> actRec == NULL)
   recArray = CreateRecArrayCopy (tree -> recOR);
 else
   recArray = CreateRecArrayCopy ((GPtrArray*) tree -> actRec -> data);
 toRet = InsertListElement (NULL, recArray);
 if (tree -> recs == NULL)
   tree -> recs = toRet;
 else
   PutAtListEnd (tree -> recs, toRet);
 ++ tree -> stored;
 return toRet;
}
